#include <config.h>
#include <assert.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <vrm.h>

void vrm_process(char *filename);

void vrm_init(void)
{
}

void vrm_done(void)
{
}

int default_callback(char *filename, char *virusname)
{
  if (virusname == NULL) {
    printf("%s is OK.\n", filename);
  } else {
    printf("%s contains %s.\n", filename, virusname);
  }
  return 0;
}

static vrm_callback callback = default_callback;

void vrm_set_callback(vrm_callback function)
{
  callback = function;
}

void vrm_process_directory(char *dirname)
{
  DIR *dir;
  struct dirent *entry;
  if ((dir = opendir(dirname))) {
    readdir(dir); /* .  */
    readdir(dir); /* .. */
    while ((entry = readdir(dir))) {
      char *name = malloc(strlen(dirname) + strlen(entry->d_name) + 2);
      assert(name != NULL);
      strcpy(name, dirname);
      if (name[strlen(name) - 1] != '/') strcat(name, "/");
      strcat(name, entry->d_name);
      vrm_process(name);
      free(name);
    }
    closedir(dir);
  } else {
    printf("Permission denied: %s\n", dirname); // FIXME: don't print here
  }
}

int infected_with_cascade(FILE *f)
{
  unsigned char jmp;
  short jmpcim;
  unsigned char entry[20];
  unsigned char seq[] = {0x02, 0xfb, 0x8c, 0xed, 0xe9, 0x01, 0x01, 0x5c, \
                         0x82, 0xec, 0x32, 0x02, 0x2f, 0xf7, 0x88, 0x2b, \
			 0x02, 0x02, 0x75, 0x10};
  int i, virusos;
  fread(&jmp, 1, 1, f);
  fread(&jmpcim, 1, 2, f);
  virusos = 0;
  if (jmp == 0xe9) {
    fseek(f, ((int)jmpcim) + 2, SEEK_SET);
    fread(entry, 1, 20, f);
    virusos = 1;
    for (i = 0; i < 20; i++) {
      if (entry[i] != seq[i] - 1) {
        virusos = 0;
      }
    }
  }
  if (virusos) return 1; else return 0;
}

void clean_cascade(char *filename)
{
  unsigned char eleje[3];
  short jmpcim;
  unsigned short xor;
  int i;
  FILE *f;
  f = fopen(filename, "rb+");
  assert (f != NULL);
  fseek(f, 1, SEEK_SET);
  fread(&jmpcim, 1, 2, f);
  jmpcim += 3;
  fseek(f, jmpcim + 0x18, SEEK_SET);
  fread(&xor, 1, 2, f);
  fseek(f, jmpcim + 0x2d, SEEK_SET);
  fread(&eleje, 1, 3, f);
  for (i = 0; i < 3; i++) {
    eleje[i] ^= (xor - 0xb - i) & 0xff;
    eleje[i] ^= ((xor - 0xa - i) & 0xff00) >> 8;
    eleje[i] ^= (jmpcim + 0x12d + i) & 0xff;
    eleje[i] ^= (jmpcim + 0x12c + i) >> 8;
  }
  fseek(f, 0, SEEK_SET);
  fwrite(&eleje, 1, 3, f);
  fflush(f);
  ftruncate(fileno(f), jmpcim - 1);
  fclose(f);
  puts("Cleaned.");
}

void vrm_process_file(char *filename)
{
  FILE *f;
  if ((f = fopen(filename, "rb"))) {
    if (infected_with_cascade(f)) {
      fclose(f);
      if (callback(filename, "virus:Cascade.1701.A")) {
        clean_cascade(filename);
      };
    } else {
      fclose(f);
      callback(filename, NULL);
    }
  } else {
    printf("Error opening %s\n", filename);  // FIXME: don't print here
  }
}

void vrm_process(char *filename)
{
  struct stat buf;
#ifdef HAVE_LSTAT
  lstat(filename, &buf);
#else
  stat(filename, &buf);
#endif
  if (S_ISDIR(buf.st_mode)) {
    vrm_process_directory(filename);
  } else {
    vrm_process_file(filename);
  }
}
